from typing import Union

import pytest
from sqlalchemy import Table, delete, insert, select, text, update
from sqlalchemy.sql.ddl import CreateTable

from pytest_capsqlalchemy.expression import SQLExpression, SQLExpressionType
from tests.conftest import Order


@pytest.mark.parametrize(
    ("sql_expression", "expected_type"),
    [
        (SQLExpression(insert(Order)), SQLExpressionType.INSERT),
        (SQLExpression(insert(Order)), "INSERT"),
        (SQLExpression(select(Order)), SQLExpressionType.SELECT),
        (SQLExpression(update(Order)), SQLExpressionType.UPDATE),
        (SQLExpression(delete(Order)), SQLExpressionType.DELETE),
        (SQLExpression(text("BEGIN")), SQLExpressionType.BEGIN),
        (SQLExpression(text("COMMIT")), SQLExpressionType.COMMIT),
        (SQLExpression(text("ROLLBACK")), SQLExpressionType.ROLLBACK),
        (SQLExpression(text("RANDOM TEXT")), SQLExpressionType.UNKNOWN),
        (SQLExpression(CreateTable(Table("some_table", Order.metadata))), SQLExpressionType.UNKNOWN),
    ],
)
def test_sql_expression_type_detection(
    sql_expression: SQLExpression, expected_type: Union[SQLExpressionType, str]
) -> None:
    assert sql_expression.type == expected_type


@pytest.mark.parametrize(
    ("sql_expression_type", "expected_is_tcl"),
    [
        (SQLExpressionType.SELECT, False),
        (SQLExpressionType.INSERT, False),
        (SQLExpressionType.UPDATE, False),
        (SQLExpressionType.DELETE, False),
        (SQLExpressionType.BEGIN, True),
        (SQLExpressionType.COMMIT, True),
        (SQLExpressionType.ROLLBACK, True),
    ],
)
def test_is_tcl_detection(sql_expression_type: SQLExpressionType, expected_is_tcl: bool) -> None:
    assert sql_expression_type.is_tcl == expected_is_tcl


@pytest.mark.parametrize(
    ("sql_expression", "bind_params", "expected_sql"),
    [
        pytest.param(
            SQLExpression(select(Order)),
            False,
            "SELECT orders.id, orders.recipient \nFROM orders",
            id="select_full_table",
        ),
        pytest.param(
            SQLExpression(select(Order.id)),
            False,
            "SELECT orders.id \nFROM orders",
            id="select_single_column",
        ),
        pytest.param(
            SQLExpression(select(Order.id).where(Order.id == 1)),
            False,
            "SELECT orders.id \nFROM orders \nWHERE orders.id = :id_1",
            id="select_where_clause_no_bind_params",
        ),
        pytest.param(
            SQLExpression(select(Order.id).where(Order.id == 1)),
            True,
            "SELECT orders.id \nFROM orders \nWHERE orders.id = 1",
            id="select_where_clause_bind_params",
        ),
        pytest.param(
            SQLExpression(insert(Order)),
            False,
            "INSERT INTO orders (id, recipient) VALUES (:id, :recipient)",
            id="test_raw_insert",
        ),
        pytest.param(
            SQLExpression(insert(Order), params={"recipient": "John Doe"}),
            False,
            "INSERT INTO orders (recipient) VALUES (:recipient)",
            id="insert_single_row_no_bind_params",
        ),
        pytest.param(
            SQLExpression(insert(Order), params={"recipient": "John Doe"}),
            True,
            "INSERT INTO orders (recipient) VALUES ('John Doe')",
            id="insert_single_row_bind_params",
        ),
        pytest.param(
            SQLExpression(insert(Order), multiparams=[{"recipient": "John Doe"}, {"recipient": "Jane Doe"}]),
            False,
            "INSERT INTO orders (recipient) VALUES (:recipient_m0), (:recipient_m1)",
            id="insert_multiple_rows_no_bind_params",
        ),
        pytest.param(
            SQLExpression(insert(Order), multiparams=[{"recipient": "John Doe"}, {"recipient": "Jane Doe"}]),
            True,
            "INSERT INTO orders (recipient) VALUES ('John Doe'), ('Jane Doe')",
            id="insert_multiple_rows_bind_params",
        ),
        pytest.param(
            SQLExpression(update(Order).where(Order.id == 1).values(recipient="John Doe")),
            False,
            "UPDATE orders SET recipient=:recipient WHERE orders.id = :id_1",
            id="update_where_clause_no_bind_params",
        ),
        pytest.param(
            SQLExpression(update(Order).where(Order.id == 1).values(recipient="John Doe")),
            True,
            "UPDATE orders SET recipient='John Doe' WHERE orders.id = 1",
            id="update_where_clause_bind_params",
        ),
        pytest.param(
            SQLExpression(delete(Order).where(Order.id == 1)),
            False,
            "DELETE FROM orders WHERE orders.id = :id_1",
            id="delete_where_clause_no_bind_params",
        ),
        pytest.param(
            SQLExpression(delete(Order).where(Order.id == 1)),
            True,
            "DELETE FROM orders WHERE orders.id = 1",
            id="delete_where_clause_bind_params",
        ),
    ],
)
def test_get_sql(sql_expression: SQLExpression, bind_params: bool, expected_sql: str) -> None:
    assert sql_expression.get_sql(bind_params=bind_params) == expected_sql



import os
from collections.abc import AsyncGenerator

import pytest
from pytest_asyncio import is_async_test
from sqlalchemy import ForeignKey
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

pytest_plugins = ["pytest_capsqlalchemy"]


def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
    pytest_asyncio_tests = (item for item in items if is_async_test(item))
    session_scope_marker = pytest.mark.asyncio(loop_scope="session")

    for async_test in pytest_asyncio_tests:
        async_test.add_marker(session_scope_marker, append=False)


@pytest.fixture(scope="session")
def db_url() -> str:
    db_name = os.environ["TEST_POSTGRES_DB"]
    db_user = os.environ["TEST_POSTGRES_USER"]
    db_password = os.environ["TEST_POSTGRES_PASSWORD"]
    db_host = os.environ["TEST_POSTGRES_HOST"]
    db_port = os.environ["TEST_POSTGRES_PORT"]

    return f"postgresql+asyncpg://{db_user}:{db_password}@{db_host}:{db_port}/{db_name}"


@pytest.fixture(scope="session")
def db_engine(db_url: str) -> AsyncEngine:
    return create_async_engine(db_url, future=True)


@pytest.fixture()
async def db_session(db_engine: AsyncEngine) -> AsyncGenerator[AsyncSession, None]:
    Session = async_sessionmaker(db_engine, class_=AsyncSession, expire_on_commit=False)

    async with Session() as s:
        yield s


# Set up models used only for testing
class TestingBaseModel(DeclarativeBase):
    __test__ = False  # Prevent pytest from trying to collect this class


class Order(TestingBaseModel):
    __tablename__ = "orders"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    recipient: Mapped[str]
    items: Mapped[list["OrderItem"]] = relationship(back_populates="order", cascade="all, delete-orphan")


class OrderItem(TestingBaseModel):
    __tablename__ = "order_items"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    item_name: Mapped[str]
    price: Mapped[float]
    order_id: Mapped[int] = mapped_column(ForeignKey("orders.id"))
    order: Mapped[Order] = relationship(back_populates="items")


@pytest.fixture(scope="session", autouse=True)
async def create_testing_tables(db_engine: AsyncEngine) -> None:
    async with db_engine.begin() as conn:
        await conn.run_sync(TestingBaseModel.metadata.create_all)

    yield

    async with db_engine.begin() as conn:
        await conn.run_sync(TestingBaseModel.metadata.drop_all)



import contextlib
import sys
from collections.abc import Mapping
from types import TracebackType
from typing import Any, Optional

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from sqlalchemy import Connection, CursorResult, Executable, text
from sqlalchemy.ext.asyncio import AsyncEngine

from pytest_capsqlalchemy.expression import SQLExpression
from pytest_capsqlalchemy.utils import temp_sqlalchemy_event


class SQLAlchemyCaptureContext:
    """
    Captures expressions executed on a SQLAlchemy engine within a specific context.
    These expressions include:

        * SELECT
        * INSERT
        * UPDATE
        * DELETE
        * BEGIN
        * COMMIT
        * ROLLBACK

    Every expression is captured as a SQLExpression object, allowing it to be parsed correctly
    and compared against.

    See SQLAlchemyCapturer for the available assertions on the captured expressions
    """

    _engine: AsyncEngine
    _captured_expressions: list[SQLExpression]

    def __init__(self, engine: AsyncEngine):
        self._engine = engine
        self._captured_expressions = []
        self._sqlaclhemy_events_stack = contextlib.ExitStack()

    @property
    def captured_expressions(self) -> list[SQLExpression]:
        return self._captured_expressions

    def clear(self) -> None:
        self._captured_expressions = []

    def _on_begin(self, conn: Connection) -> None:
        self._captured_expressions.append(SQLExpression(executable=text("BEGIN")))

    def _on_commit(self, conn: Connection) -> None:
        self._captured_expressions.append(SQLExpression(executable=text("COMMIT")))

    def _on_rollback(self, conn: Connection) -> None:
        self._captured_expressions.append(SQLExpression(executable=text("ROLLBACK")))

    def _on_after_execute(
        self,
        conn: Connection,
        clauseelement: Executable,
        multiparams: list[dict[str, Any]],
        params: dict[str, Any],
        execution_options: Mapping[str, Any],
        result: CursorResult,
    ) -> None:
        self._captured_expressions.append(
            SQLExpression(executable=clauseelement, params=params, multiparams=multiparams)
        )

    def __enter__(self) -> Self:
        events_stack = self._sqlaclhemy_events_stack.__enter__()

        for event_name, listener in (
            ("begin", self._on_begin),
            ("commit", self._on_commit),
            ("rollback", self._on_rollback),
            ("after_execute", self._on_after_execute),
        ):
            events_stack.enter_context(
                temp_sqlalchemy_event(
                    self._engine.sync_engine,
                    event_name,
                    listener,
                )
            )

        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        return self._sqlaclhemy_events_stack.__exit__(exc_type, exc_value, traceback)



import sys
from types import TracebackType
from typing import Optional, Union

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from sqlalchemy.ext.asyncio import AsyncEngine

from pytest_capsqlalchemy.context import SQLAlchemyCaptureContext
from pytest_capsqlalchemy.expression import SQLExpression, SQLExpressionType


class SQLAlchemyCapturer:
    """
    The main fixture class for the `capsqlalchemy` plugin. Used to perform asserts
    about the expressions SQLAlchemy has executed during the test.

    Can be used either directly using the assert methods (to perform checks for all expressions
    executed in the test), as a context manager (to perform checks only for the expressions
    executed in a specific block), or a combination of both.

    Intended to be used via the [`capsqlalchemy`][pytest_capsqlalchemy.plugin.capsqlalchemy] fixture
    """

    _full_test_context: SQLAlchemyCaptureContext
    _partial_context: Optional[SQLAlchemyCaptureContext]

    def __init__(self, full_test_context: SQLAlchemyCaptureContext):
        self._full_test_context = full_test_context
        self._partial_context = None

    @property
    def engine(self) -> AsyncEngine:
        return self._full_test_context._engine

    @property
    def captured_expressions(self) -> list[SQLExpression]:
        """
        Returns all SQL expressions captured in the current context.

        When used outside a context manager block, returns all expressions captured
        during the entire test. When used inside a context manager block, returns
        only the expressions captured within that specific block.

        This property is useful for performing specific assertions on the captured expressions which
        cannot be easily achieved with the provided assert methods.
        """

        if self._partial_context is not None:
            return self._partial_context.captured_expressions

        return self._full_test_context.captured_expressions

    def __enter__(self) -> Self:
        self._partial_context = SQLAlchemyCaptureContext(self.engine)
        self._partial_context = self._partial_context.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exc_tb: Optional[TracebackType] = None,
    ) -> Optional[bool]:
        if self._partial_context is None:  # pragma: no cover
            raise RuntimeError(f"{self.__class__.__name__}: attempting to call __exit__ before __enter__")

        result = self._partial_context.__exit__(exc_type, exc_val, exc_tb)

        self._partial_context = None

        return result

    def assert_query_types(
        self,
        *expected_query_types: Union[SQLExpressionType, str],
        include_tcl: bool = True,
    ) -> None:
        """
        Asserts that the captured SQL expressions match the expected query types in order.
        This is useful for ensuring that your code is generating correct query types but
        their exact strcutre is not important (e.g. complex SELECT statements).

        Args:
            *expected_query_types: Variable number of expected query types
            include_tcl: Whether to include transaction control language statements (BEGIN,
                COMMIT, ROLLBACK) in the comparison

        Raises:
            AssertionError: If the actual query types don't match the expected ones.
        """

        actual_query_types_values = []

        for query in self.captured_expressions:
            if not include_tcl and query.type.is_tcl:
                continue

            actual_query_types_values.append(query.type._value_)

        # Converting to strings as the error message diff will be shorter and more readable
        expected_query_types_values = [
            query_type._value_ if isinstance(query_type, SQLExpressionType) else query_type
            for query_type in expected_query_types
        ]

        assert expected_query_types_values == actual_query_types_values

    def assert_query_count(self, expected_query_count: int, *, include_tcl: bool = True) -> None:
        """
        Asserts that the number of captured SQL expressions matches the expected count.
        This is useful for ensuring that your code is not generating more statements than expected
        (e.g. due to N+1 queries), however the exact queries are not important.

        Args:
            expected_query_count: The expected number of SQL expressions.
            include_tcl: Whether to include transaction control language statements (BEGIN,
                COMMIT, ROLLBACK) in the count.

        Raises:
            AssertionError: If the actual query count doesn't match the expected count.
        """
        actual_query_count = 0

        for query in self.captured_expressions:
            if not include_tcl and query.type.is_tcl:
                continue

            actual_query_count += 1

        assert expected_query_count == actual_query_count, (
            f"Query count mismatch: expected {expected_query_count}, got {actual_query_count}"
        )

    def assert_captured_queries(
        self,
        *expected_queries: str,
        include_tcl: bool = True,
        bind_params: bool = False,
    ) -> None:
        """
        Asserts that the captured SQL queries match the expected SQL strings in order.
        This is useful for ensuring that your code is generating the exact SQL statements you expect.

        Args:
            *expected_queries: Variable number of expected SQL query strings.
            include_tcl: Whether to include transaction control language statements (BEGIN,
                COMMIT, ROLLBACK) in the comparison.
            bind_params: Whether to include bound parameters in the SQL strings. When `False`,
                parameters are represented as placeholders instead.

        Raises:
            AssertionError: If the actual SQL queries don't match the expected ones.
        """
        actual_queries = []

        for query in self.captured_expressions:
            if not include_tcl and query.type.is_tcl:
                continue

            actual_queries.append(query.get_sql(bind_params=bind_params))

        assert list(expected_queries) == actual_queries



from sqlalchemy import select, text
from sqlalchemy.ext.asyncio import AsyncSession

from pytest_capsqlalchemy.context import SQLAlchemyCaptureContext
from tests.conftest import Order


async def test_capture_session_simple_select(
    capsqlalchemy_context: SQLAlchemyCaptureContext,
    db_session: AsyncSession,
) -> None:
    await db_session.execute(select(text("1")))

    expr_types = [expr.type._value_ for expr in capsqlalchemy_context.captured_expressions]
    assert expr_types == ["BEGIN", "SELECT"]


async def test_capture_session_select_in_explicit_transaction(
    capsqlalchemy_context: SQLAlchemyCaptureContext,
    db_session: AsyncSession,
) -> None:
    async with db_session.begin():
        await db_session.execute(select(text("1")))

    expr_types = [expr.type._value_ for expr in capsqlalchemy_context.captured_expressions]
    assert expr_types == ["BEGIN", "SELECT", "COMMIT"]


async def test_capture_session_mixed_select_statements_multi_transactions(
    capsqlalchemy_context: SQLAlchemyCaptureContext,
    db_session: AsyncSession,
) -> None:
    await db_session.execute(select(text("1")))

    await db_session.rollback()

    async with db_session.begin():
        await db_session.execute(select(text("1")))
        await db_session.execute(select(text("1")))

    await db_session.execute(select(text("1")))
    await db_session.commit()
    await db_session.execute(select(text("1")))

    expr_types = [expr.type._value_ for expr in capsqlalchemy_context.captured_expressions]
    assert expr_types == [
        "BEGIN",
        "SELECT",
        "ROLLBACK",
        "BEGIN",
        "SELECT",
        "SELECT",
        "COMMIT",
        "BEGIN",
        "SELECT",
        "COMMIT",
        "BEGIN",
        "SELECT",
    ]


async def test_clear(
    capsqlalchemy_context: SQLAlchemyCaptureContext,
    db_session: AsyncSession,
) -> None:
    await db_session.execute(select(text("1")))
    await db_session.execute(select(text("1")))

    capsqlalchemy_context.clear()

    await db_session.execute(select(text("1")))

    assert len(capsqlalchemy_context.captured_expressions) == 1
    assert capsqlalchemy_context.captured_expressions[0].type == "SELECT"


async def test_capture_session_add_single_order(
    capsqlalchemy_context: SQLAlchemyCaptureContext,
    db_session: AsyncSession,
) -> None:
    db_session.add(Order(recipient="John Doe"))
    await db_session.commit()

    expr_types = [expr.type._value_ for expr in capsqlalchemy_context.captured_expressions]
    assert expr_types == ["BEGIN", "INSERT", "COMMIT"]

    insert_expr = capsqlalchemy_context.captured_expressions[1]
    assert insert_expr.params == {"recipient": "John Doe"}
    assert insert_expr.multiparams == []


async def test_capture_session_add_multiple_orders(
    capsqlalchemy_context: SQLAlchemyCaptureContext,
    db_session: AsyncSession,
) -> None:
    db_session.add(Order(recipient="John Doe"))
    db_session.add(Order(recipient="Jane Doe"))
    await db_session.commit()

    expr_types = [expr.type._value_ for expr in capsqlalchemy_context.captured_expressions]
    assert expr_types == ["BEGIN", "INSERT", "COMMIT"]

    insert_expr = capsqlalchemy_context.captured_expressions[1]
    assert insert_expr.params == {}
    assert insert_expr.multiparams == [{"recipient": "John Doe"}, {"recipient": "Jane Doe"}]


async def test_capture_session_update_orders(
    capsqlalchemy_context: SQLAlchemyCaptureContext,
    db_session: AsyncSession,
) -> None:
    order = Order(recipient="John Doe")
    db_session.add(order)
    await db_session.commit()

    order.recipient = "Jane Doe"
    await db_session.commit()

    update_expr = next(expr for expr in capsqlalchemy_context.captured_expressions if expr.type == "UPDATE")
    assert update_expr.params == {
        "orders_id": order.id,
        "recipient": "Jane Doe",
    }
    assert update_expr.multiparams == []



from sqlalchemy import select, text
from sqlalchemy.ext.asyncio import AsyncSession

from pytest_capsqlalchemy import SQLAlchemyCapturer
from pytest_capsqlalchemy.expression import SQLExpressionType
from tests.conftest import Order, OrderItem


async def test_assert_query_count(db_session: AsyncSession, capsqlalchemy: SQLAlchemyCapturer) -> None:
    await db_session.execute(text("SELECT 1"))
    await db_session.execute(select(OrderItem))

    await db_session.commit()

    async with db_session.begin():
        await db_session.execute(select(Order).where(Order.id == 1))
        await db_session.execute(select(text("1")))

        db_session.add(Order(recipient="John Doe"))

    capsqlalchemy.assert_query_count(9, include_tcl=True)
    capsqlalchemy.assert_query_count(5, include_tcl=False)


async def test_changing_context(db_session: AsyncSession, capsqlalchemy: SQLAlchemyCapturer) -> None:
    await db_session.execute(text("SELECT 1"))

    capsqlalchemy.assert_query_count(1, include_tcl=False)

    await db_session.execute(text("SELECT 1"))

    capsqlalchemy.assert_query_count(2, include_tcl=False)

    with capsqlalchemy:
        capsqlalchemy.assert_query_count(0, include_tcl=False)
        await db_session.execute(text("SELECT 1"))
        capsqlalchemy.assert_query_count(1, include_tcl=False)

    capsqlalchemy.assert_query_count(3, include_tcl=False)


async def test_captured_queries(db_session: AsyncSession, capsqlalchemy: SQLAlchemyCapturer) -> None:
    await db_session.execute(text("SELECT 1"))
    await db_session.execute(select(OrderItem))

    await db_session.commit()

    async with db_session.begin():
        await db_session.execute(select(Order).where(Order.id == 1))
        await db_session.execute(select(text("1")))

        db_session.add(Order(recipient="John Doe"))

    capsqlalchemy.assert_captured_queries(
        "BEGIN",
        "SELECT 1",
        "SELECT order_items.id, order_items.item_name, order_items.price, order_items.order_id \nFROM order_items",
        "COMMIT",
        "BEGIN",
        "SELECT orders.id, orders.recipient \nFROM orders \nWHERE orders.id = :id_1",
        "SELECT 1",
        "INSERT INTO orders (recipient) VALUES (:recipient)",
        "COMMIT",
    )

    capsqlalchemy.assert_captured_queries(
        "SELECT 1",
        "SELECT order_items.id, order_items.item_name, order_items.price, order_items.order_id \nFROM order_items",
        "SELECT orders.id, orders.recipient \nFROM orders \nWHERE orders.id = :id_1",
        "SELECT 1",
        "INSERT INTO orders (recipient) VALUES (:recipient)",
        include_tcl=False,
    )


async def test_captured_queries_bind_params(db_session: AsyncSession, capsqlalchemy: SQLAlchemyCapturer) -> None:
    await db_session.execute(text("SELECT 1"))
    await db_session.execute(select(OrderItem))

    await db_session.commit()

    async with db_session.begin():
        await db_session.execute(select(Order).where(Order.id == 1))
        await db_session.execute(select(text("1")))

        db_session.add(Order(recipient="John Doe"))

    capsqlalchemy.assert_captured_queries(
        "BEGIN",
        "SELECT 1",
        "SELECT order_items.id, order_items.item_name, order_items.price, order_items.order_id \nFROM order_items",
        "COMMIT",
        "BEGIN",
        "SELECT orders.id, orders.recipient \nFROM orders \nWHERE orders.id = 1",
        "SELECT 1",
        "INSERT INTO orders (recipient) VALUES ('John Doe')",
        "COMMIT",
        bind_params=True,
    )


async def test_captured_query_types(db_session: AsyncSession, capsqlalchemy: SQLAlchemyCapturer) -> None:
    await db_session.execute(select(OrderItem))

    await db_session.commit()

    async with db_session.begin():
        await db_session.execute(select(Order).where(Order.id == 1))
        await db_session.execute(select(text("1")))

        db_session.add(Order(recipient="John Doe"))

    capsqlalchemy.assert_query_types(
        "BEGIN",
        "SELECT",
        "COMMIT",
        "BEGIN",
        "SELECT",
        "SELECT",
        "INSERT",
        "COMMIT",
    )

    capsqlalchemy.assert_query_types(
        SQLExpressionType.SELECT,
        SQLExpressionType.SELECT,
        SQLExpressionType.SELECT,
        SQLExpressionType.INSERT,
        include_tcl=False,
    )


async def test_captured_queries_insert_with_relationship(
    db_session: AsyncSession, capsqlalchemy: SQLAlchemyCapturer
) -> None:
    async with db_session.begin():
        order = Order(recipient="John Doe")
        db_session.add(order)
        db_session.add(OrderItem(item_name="Bread", price=2.00, order=order))
        db_session.add(OrderItem(item_name="Butter", price=3.50, order=order))

    capsqlalchemy.assert_captured_queries(
        "INSERT INTO orders (recipient) VALUES (:recipient)",
        (
            "INSERT INTO order_items (item_name, price, order_id) VALUES "
            "(:item_name_m0, :price_m0, :order_id_m0), "
            "(:item_name_m1, :price_m1, :order_id_m1)"
        ),
        include_tcl=False,
    )

    capsqlalchemy.assert_captured_queries(
        "INSERT INTO orders (recipient) VALUES ('John Doe')",
        (
            "INSERT INTO order_items (item_name, price, order_id) VALUES "  # noqa: S608
            f"('Bread', 2.0, {order.id}), "
            f"('Butter', 3.5, {order.id})"
        ),
        include_tcl=False,
        bind_params=True,
    )



import enum
from dataclasses import dataclass, field
from typing import Any

from sqlalchemy import ClauseElement, Executable, Insert, TextClause, text


class SQLExpressionType(str, enum.Enum):
    SELECT = "SELECT"
    INSERT = "INSERT"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    BEGIN = "BEGIN"
    COMMIT = "COMMIT"
    ROLLBACK = "ROLLBACK"
    UNKNOWN = "UNKNOWN"

    @property
    def is_tcl(self) -> bool:
        return self in {SQLExpressionType.BEGIN, SQLExpressionType.COMMIT, SQLExpressionType.ROLLBACK}


@dataclass
class SQLExpression:
    executable: Executable
    params: dict[str, Any] = field(default_factory=dict)
    multiparams: list[dict[str, Any]] = field(default_factory=list)

    def get_sql(self, *, bind_params: bool = False) -> str:
        assert isinstance(self.executable, ClauseElement)

        if self.executable.is_insert:
            assert isinstance(self.executable, Insert)

            if self.multiparams:
                expr = self.executable.values(self.multiparams)
            elif self.params:
                expr = self.executable.values(self.params)
            else:
                expr = self.executable
        else:
            expr = self.executable

        compile_kwargs = {}
        if bind_params:
            compile_kwargs["literal_binds"] = True

        return str(expr.compile(compile_kwargs=compile_kwargs))

    @property
    def type(self) -> SQLExpressionType:
        if self.executable.is_insert:
            return SQLExpressionType.INSERT

        if self.executable.is_select:
            return SQLExpressionType.SELECT

        if self.executable.is_update:
            return SQLExpressionType.UPDATE

        if self.executable.is_delete:
            return SQLExpressionType.DELETE

        if isinstance(self.executable, TextClause):
            if self.executable.compare(text("BEGIN")):
                return SQLExpressionType.BEGIN

            if self.executable.compare(text("COMMIT")):
                return SQLExpressionType.COMMIT

            if self.executable.compare(text("ROLLBACK")):
                return SQLExpressionType.ROLLBACK

        return SQLExpressionType.UNKNOWN



from pytest_capsqlalchemy.capturer import SQLAlchemyCapturer
from pytest_capsqlalchemy.context import SQLAlchemyCaptureContext
from pytest_capsqlalchemy.expression import SQLExpression
from pytest_capsqlalchemy.plugin import capsqlalchemy, capsqlalchemy_context

__all__ = [
    "SQLAlchemyCaptureContext",
    "SQLAlchemyCapturer",
    "SQLExpression",
    "capsqlalchemy",
    "capsqlalchemy_context",
]



import contextlib
from collections.abc import Callable, Generator
from typing import Any

from sqlalchemy import event


@contextlib.contextmanager
def temp_sqlalchemy_event(
    target: Any,
    identifier: str,
    fn: Callable[..., Any],
    *args: Any,
    **kwargs: Any,
) -> Generator[None, None, None]:
    event.listen(target, identifier, fn, *args, **kwargs)

    try:
        yield
    finally:
        event.remove(target, identifier, fn)



from collections.abc import Generator

import pytest
from sqlalchemy.ext.asyncio import AsyncEngine

from pytest_capsqlalchemy import SQLAlchemyCaptureContext, SQLAlchemyCapturer


@pytest.fixture
def capsqlalchemy_context(db_engine: AsyncEngine) -> Generator[SQLAlchemyCaptureContext]:
    with SQLAlchemyCaptureContext(db_engine) as capsqlalchemy_ctx:
        yield capsqlalchemy_ctx


@pytest.fixture()
def capsqlalchemy(capsqlalchemy_context: SQLAlchemyCaptureContext) -> SQLAlchemyCapturer:
    """The main fixture!"""
    return SQLAlchemyCapturer(capsqlalchemy_context)
